/*
	majamemmgr.h
	ce fichier est une partie de Maja - this is a part of Maja
	gestion principale de la mmoire
	main memory management

	$Id: $

	$Log: $

	begin:		Jan 2003
	copyright:	(C) 2000,2001,2002,2003 by Jean-Dominique Frattini
	email:		jd.frattini@tuxfamily.org	
*/

/************************************************************************************************
 *												*
 *	Ce programme est un logiciel libre; vous pouvez le redistribuer et/ou le modifier	*
 *	sous les termes de la GNU General Public License comme publie par			*
 *	la Free Software Foundation; soit en utilisant la verstion 2 de cette license, ou	*
 *	(si vous le voulez) toute version ultrieure.						*
 *	-											*
 *	This program is free software; you can redistribute it and/or modify			*
 *	it under the terms of the GNU General Public License as published by			*
 *	the Free Software Foundation; either version 2 of the License, or			*
 *	(at your option) any later version.							*
 *												*
 ************************************************************************************************/

#ifndef __MAJA_MEMMGR_H
#define __MAJA_MEMMGR_H

#include <iostream>
#include "../common/majatypes.h"
#include "../engines/graphic/graphext.h"

namespace maja
{
	/*!
	 * memory type to be used with the
	 * memory manager.
	 */
	enum mtype{
		//! none: there is no memory set (never use it !)
		none_t= 0,
		
		//! main: this is the system memory (RAM)
		//! this memory is available for all plateforms.
		main_t= 1,

		//! agp: this is one of system AGP or graphic AGP
		//! memory. Modify your priority to get more chance
		//! to be in graphic memory. This is only available
		//! for NVidia AGP cards (since TNT).
		agp_t = 2,
		
		//! both: use main and agp memories together.
		both_t= 4
	};

	/*!
	 * manage the system memory of maja.
	 * this was created due to a problem found with glXAllocateMemoryNV
	 * in multi-threading mode: we could only allocate from the main
	 * thread, that is not understandable as this is written in the referring
	 * doc I read. cf: http://oss.sgi.com/projects/ogl-sample/registry/NV/vertex_array_range.txt
	 * this is the official opengl extensions doc.
	 * Use it whenever you need some a huge amount of memory to be 'well managed', and assuring
	 * that no corrupt is welcome. Use it from the main thread (in fact, best in system or
	 * scheduler), whenever you need to allocate AGP memory.
	 * This implies user has to know at start the amount of memory the application requires. This
	 * is as if we know before what will be in the game. So next implementation modifications
	 * will round this issue by giving a singleton of this class, shared by all the threads.
	 * This way, we will able to review the memory we wanted.
	 * 
	 * infos: this is more intended to allocate primitive types as float, int... more
	 * 	than using complex types as classes.
	 *
	 * status: useable carefuly, not til a user class
	 */
	template <typename T= mfloat>
	class MajaMemoryManager
	{
		protected:
		
		mulong		main_memory_size;
		mulong		main_memory_used;
		mulong		agp_memory_size;
		mulong		agp_memory_used;

		mbool		agp_is_not_agp;

		MajaGExt	g_ext;

		mfloat		rf;
		mfloat		wf;
		mfloat		pv;

		T*		p_agp_memory_start;
		T*		p_main_memory_start;
		T*		p_agp_memory_free;
		T*		p_main_memory_free;

		mulong		num_agp_blocks;
		mulong		num_main_blocks;

		mtype		type;
		mbool		mem_init;

		public:

		MajaMemoryManager( mvoid);

		protected:

		MajaMemoryManager( const MajaMemoryManager<>& mmm);

		public:

		~MajaMemoryManager( mvoid);

		//! call it after each instanciation in order to initialize
		//! the memory management. You indicates here the maximum size you want to allocate
		//! in the aimed memories, and each call to future Allocate() will give you a part
		//! of this allocation. So, just remember you consider here the maximum size you will
		//! end it up with it.
		//! main_size is the size in bits of the requested size in the RAM
		//! agp_size is the size in bits of the requested size in the AGP memory (VRAM or not)
		//! initializes internally the memory type referring to whatever we give to args.
		mvoid InitMemory( mulong main_size= 0, mulong agp_size= 0);

		protected:

		mvoid Init( mvoid);

		mvoid SetMemory( mvoid);

		public:

		mtype SetMemoryType( mtype t);
		mtype GetMemoryType( mvoid);

		//! call it when the object is fully initialized only ! (when MemInit() returns)
		//! in order to get a requested portion (or all) of what you allocated before.
		//! you could not allocate to both at the same time, this will result in a fail
		//! by returning mnull.
		//! if only one memory is available, always use this memory, with no regard to
		//! what you give for mtype mt.
		mvoid* Allocate( mlong size, mtype mt= none_t);

		mvoid Free( mvoid* ptr, mtype mt= none_t);

		mvoid PrintInfo( mvoid);
	};

	// this specialization (or maybe the generic one) looks like a shame redonduncy...
	template <>
	class MajaMemoryManager<mfloat>
	{
		protected:
		
		mulong		main_memory_size;
		mulong		main_memory_used;
		mulong		agp_memory_size;
		mulong		agp_memory_used;

		mbool		agp_is_not_agp;

		MajaGExt	g_ext;

		mfloat		rf;
		mfloat		wf;
		mfloat		pv;

		mfloat*		p_agp_memory_start;
		mfloat*		p_main_memory_start;
		mfloat*		p_agp_memory_free;
		mfloat*		p_main_memory_free;

		mulong		num_agp_blocks;
		mulong		num_main_blocks;

		mtype		type;
		mbool		mem_init;
		
		public:

		MajaMemoryManager( mvoid);

		protected:

		MajaMemoryManager( const MajaMemoryManager<>& mmm);

		public:

		~MajaMemoryManager( mvoid);

		//! call it after each instanciation in order to initialize
		//! the memory management. You indicates here the maximum size you want to allocate
		//! in the aimed memories, and each call to future Allocate() will give you a part
		//! of this allocation. So, just remember you consider here the maximum size you will
		//! end it up with it.
		//! main_size is the size in bits of the requested size in the RAM
		//! agp_size is the size in bits of the requested size in the AGP memory (VRAM or not)
		//! initializes internally the memory type referring to whatever we give to args.
		mvoid InitMemory( mulong main_size= 0, mulong agp_size= 0);

		protected:

		mvoid Init( mvoid);

		mvoid SetMemory( mvoid);

		public:

		mtype SetMemoryType( mtype t);
		mtype GetMemoryType( mvoid);

		//! call it when the object is fully initialized only ! (when MemInit() returns)
		//! in order to get a requested portion (or all) of what you allocated before.
		//! you could not allocate to both at the same time, this will result in a fail
		//! by returning mnull.
		//! if only one memory is available, always use this memory, with no regard to
		//! what you give for mtype mt.
		mvoid* Allocate( mlong size, mtype mt= none_t);

		mvoid Free( mvoid* ptr, mtype mt= none_t);

		mvoid PrintInfo( mvoid);
	};

}//namespace maja
#endif
