/*
	graphicmodule.h
	ce fichier est une partie de Maja - this is a part of Maja
	le module graphique
	the graphic module
	
	$Id: $

	$Log: $

	begin:			Fri Jul 5 2002
	copyright:		(C) 2000,2001,2002,2003 by Jean-Dominique Frattini
	email:			jd.frattini@tuxfamily.org	
*/

/************************************************************************************************
 *												*
 *	Ce programme est un logiciel libre; vous pouvez le redistribuer et/ou le modifier	*
 *	sous les termes de la GNU General Public License comme publie par			*
 *	la Free Software Foundation; soit en utilisant la verstion 2 de cette license, ou	*
 *	(si vous le voulez) toute version ultrieure.						*
 *	-											*
 *	This program is free software; you can redistribute it and/or modify			*
 *	it under the terms of the GNU General Public License as published by			*
 *	the Free Software Foundation; either version 2 of the License, or			*
 *	(at your option) any later version.							*
 *												*
 ************************************************************************************************/

/*! \file
	Ordonnancement des diffrents modules que peut comporter Maja
	Scheduling of the different modules that can stand Maja
*/

#ifndef __MAJA_GRAPHICMODULE_H
#define __MAJA_GRAPHICMODULE_H

#include <typeinfo>
#include "../../core/majascheduler.h"
#include "../../core/majasmodule.h"
#include "../../engines/graphic/majadisplay.h"	// think about not knowing what is the graphic in real !!
#include "../../layers/th/mutex.h"
#include "../../layers/wm/majaxwindow.h"

namespace maja
{
#ifdef SHOW_SRENDER_INFO
#undef SHOW_SRENDER_INFO
#endif

// uncomment it to show this module's infos
//#define SHOW_SRENDER_INFO

namespace MAJAGRAPHIC
{
using namespace MAJACORE;

//! this is mtrue if we receive any event meaning close the window
static bool maja_window_close= mfalse;

//! this below tries to behave like policy-based classes

/*! 
 * Maja has different ways to render, and the user has to tell her
 * which one to follow. Users could also make their own rendering
 * ways for any reasons. So, the library provides two main ways and
 * so, the user could behave in front of this. 
 * One choose the MajaDisplay (not to be confused with the windowing
 * _MajaDisplay type/member) and the other choose MajaWindow to render.
 * 
 * This class is intended for the first policy case above and has some
 * behaviors that restrict a bit its ways but give her the power of using
 * internally the graphic engine.
 * Users have to fill MajaDisplay with MajaObjects polymorphic objects
 * and a camera in order to get a correct rendering context.
 * This is the recommanded way for rendering in Maja.
 *
 * This is the default way maja is rendering. It needs two other classes
 * as templates:
 * _D is the rendering engine (since now MajaDisplay),
 * _C is the opengl context.
 *
 * This will generally be called the maja renderer.
 * 
 * status: useable
 */
template 
<
	class _D=MajaDisplay,
	class _C=_MajaContext
>
struct MajaMethod_Display
{
	_D *_d;
	volatile _C _c;

	mint Draw( mvoid)
	{
		return _d->Draw();
	}

	volatile _MajaContext CreateContext( mvoid)
	{
		return _c=_d->GetWindow().CreateContext();
	}

	volatile mvoid MakeContextCurrent( _C c=mnull)
	{
		_d->GetDrawable().Enter( c?c:(_C)(_c));
	}

	mvoid LeaveCurrentContext( mvoid){
		_d->GetDrawable().Leave();
	}

	MajaMethod_Display( _D *disp=mnull)
	{
		_d=disp;
	}
};

/*!
 * This is the window policy rendering way.
 * Here, users won't attempt Maja to do this stuff, but
 * more behave as an opengl overlay (a bit same as glut)
 * in the way that it will need to derived from a
 * MajaWindow hierarchy class and to overload some of its
 * member functions (please cf layers/wm for more information
 * about that).
 *
 * this will generally be called the maja renderable.
 * 
 * status: need some more tests (efficiency, maja whole compatibility...)
 */
template <class _D=MajaWindow>
struct MajaMethod_Window
{
	_D *_d;
	_MajaContext _c;

	mint Draw( mvoid)
	{
		_d->Draw();
		return 1;
	}

	_MajaContext CreateContext( mvoid)
	{
		return _c= glXCreateContext( _d->GetDisplay(), _d->GetVisual(), 0, GL_TRUE);
	}

	mvoid MakeContextCurrent( mvoid)
	{
		glXMakeCurrent(	_d->GetDisplay(), _d->GetWindow(), _c);
	}

	MajaMethod_Window( _D *win=mnull)
	{
		_d=win;
	}
};

/*!
 * A simple threaded module for managing the renderer.
 * You will need to define its template argument with
 * one rendering policy (use one of them above or just
 * create your own).
 * Users should never use it.
 *
 * status: useable.
 */
template 
<
	class MajaRenderPolicy
>
class MajaSRender:	public MajaSModule, 
			public MajaRenderPolicy
{
	public:
		
	MajaSRender( mvoid):	MajaSModule(), 
				MajaRenderPolicy()
	{
		maja_schedule_algo= SCHED_FIFO;
		maja_window_close= mfalse;
	}

	~MajaSRender( mvoid){
		;
	}
	
	/*!
	 * get enter the thread execution
	 * you could give here parameters to give to the thread
	 */
	mint EnterDefaultMode( mvoid *params= mnull){

		sched_param sp;

		GetPriority();
		
		pthread_attr_init( &attributes);
		pthread_attr_setschedpolicy( &attributes, maja_schedule_algo);
		pthread_attr_getschedparam( &attributes, &sp);		
		sp.sched_priority= priority;
		pthread_attr_setschedparam( &attributes, &sp);

		if( !Create( mnull)){
			std::cerr << std::endl << "could not create MajaSRender module thread";
			if( errno==EPERM)
				std::cerr << std::endl << "you may not be seteuid=0";

			return 0;
		}

		return 1;
	}

	mvoid* Go( mvoid* args)
	{
		mbool			done=			mfalse;
		mint			last_cancel_state=	0;
		_MajaContext		context= mnull;

		
		pthread_setcancelstate( PTHREAD_CANCEL_DISABLE, &last_cancel_state);

		Lock();
		if( GetState()!=MMS_state_start){
			maja_core_mutex.stderr_mutex.Lock();
			std::cerr << std::endl << "MajaSRender: wanted to go, but not started ! exiting...";
			maja_core_mutex.stderr_mutex.Unlock();
			SetState( MMS_state_off);
			Unlock();

			return mnull;
		}
		Unlock();

		maja_core_mutex.stderr_mutex.Lock();
		std::cerr << std::endl << "MajaSRender: waiting for scheduler order...";
		maja_core_mutex.stderr_mutex.Unlock();

		Lock();
		while( state!=MMS_state_on)
			CondWait();
		Unlock();
		
		maja_core_mutex.stderr_mutex.Lock();
		std::cerr << std::endl << "MajaSRender: creating renderable...";
		
		Lock();		
		_d->GetWindow().Lock();
		_d->GetWindow().GetDrawable().Lock();
		context= _d->GetWindow().GetDrawable().CreateContext();
		_d->GetWindow().GetDrawable().Unlock();
		_d->GetWindow().Unlock();
		Unlock();
		
		std::cerr << " done." << std::flush;
		maja_core_mutex.stderr_mutex.Unlock();

		while( mtrue){
			Lock();
			_d->Lock();
			_d->GetWindow().Lock();
			if( _d->GetComputed()){
				_d->GetWindow().GetDrawable().Lock();
				MakeContextCurrent( context);
				_d->GetGEngine().Compute();
				_d->GetGEngine().Check();
				_d->GetWindow().GetDrawable().Unlock();
				_d->GetWindow().Unlock();
				_d->Unlock();
				Unlock();
				break;
			}
			_d->GetWindow().Unlock();
			_d->Unlock();
			Unlock();
			usleep( 10000);
		}

		mint rr=0;
		
		std::cerr << std::endl << "MajaSRender: rounding...";
		while( !done){
			Lock();
			switch( GetState()){			
				case MMS_state_on:
				case MMS_state_busy:
					break;
				case MMS_state_pause:
					continue;
				case MMS_state_cancel:
					done= mtrue;
					break;
				case MMS_state_off:
				default:
					maja_core_mutex.stderr_mutex.Lock();
					std::cerr << std::endl << "exiting from MajaSRender::Go()";
					maja_core_mutex.stderr_mutex.Unlock();	
					Unlock();
					pthread_exit( mnull);
			}
			
			SetState( MMS_state_busy);		
			_d->GetWindow().GetDrawable().Lock();
			MakeContextCurrent( context);
			rr= _d->Draw();
			LeaveCurrentContext();
			_d->GetWindow().GetDrawable().Unlock();

			if( rr== -1){
				done= true;
			}
			
			if( GetState()!=MMS_state_off)
				SetState( MMS_state_on);
			else{
				Unlock();
				break;
			}
			Unlock();
	
		} // while

			
		pthread_exit( mnull);
	}
};

//! MajaRendererModule is the main node for using MajaDisplay as the drawer
typedef MajaSRender<MajaMethod_Display<> > MajaRendererModule;

//! MajaRenderableModule is the main node for using MajaWindow as the drawer
typedef MajaSRender<MajaMethod_Window<> > MajaRenderableModule;

}
}
#endif
