/*
	majagameengine.h
	ce fichier est une partie de Maja - this is a part of Maja
	le moteur de jeux
	the game engine

	$Id: $

	$Log: $

	begin:			nov 2002
	copyright:		(C) 2000,2001,2002,2003 by Jean-Dominique Frattini
	email:			jd.frattini@tuxfamily.org	
*/

/************************************************************************************************
 *												*
 *	Ce programme est un logiciel libre; vous pouvez le redistribuer et/ou le modifier	*
 *	sous les termes de la GNU General Public License comme publie par			*
 *	la Free Software Foundation; soit en utilisant la verstion 2 de cette license, ou	*
 *	(si vous le voulez) toute version ultrieure.						*
 *	-											*
 *	This program is free software; you can redistribute it and/or modify			*
 *	it under the terms of the GNU General Public License as published by			*
 *	the Free Software Foundation; either version 2 of the License, or			*
 *	(at your option) any later version.							*
 *												*
 ************************************************************************************************/

#ifndef MAJA_GAMEENGINE__H
#define MAJA_GAMEENGINE__H

#include "majagamemenu.h"
#include "majagameworld.h"
#include "majagameloader.h"
#include "../../common/majacammodes.h"

namespace maja
{

/*!file
 *
 * The game engine is a young machine, so don't expect it
 * to be as implemented as some other parts are.
 * Not all the things work, they're here just for concept.
 */
	
/*!
 * This is the main state of the game that evaluates the main
 * living way of the engine.
 */
enum MajaGameState
{
	game_state_off= 0,
	game_state_starting,
	game_state_loading,
	game_state_living
};

/*!
 * the living sub states.
 * It's possible that the state is a combination
 * of some of them, but it's important to get sense,
 * so off is disabled whenever another state is set,
 * and pause hide all others.
 */
typedef muint MajaLivingState;

//! there is no running due to this state, so this
//! state may never exists but when starting.
static const muint living_state_off	= 0x0;

//! the game enter a menu
static const muint living_state_menuing	= 0x1;

//! the game enter the game
static const muint living_state_living	= 0x2;

//! the game enter the hud
static const muint living_state_huding	= 0x4;

//! the game hangs in a pause
static const muint living_state_pause	= 0x8;

/*!
 * namespace containing all for gaming
 */
namespace MAJAGAME
{
	/*!
	 * Manages the rounding of the game.
	 * This is a very simple engine that is only at its beginning, so it
	 * lacks many things that could be important.
	 * MajaGameEngine_ is a states machine, and its behave depends on them.
	 * 
	 * O is any MajaObject
	 * I is the interactive engine (MajaIEngine)
	 * S is the scenery engine (MajaSEngine)
	 * R is the rendering engine (at this time it's MajaDisplay)
	 * M is the menu engine (MajaMenu)
	 * T is the texture loader (MajaTextures)
	 */
	template 
	<
		class O, 
		class I, 
		class S,
		class R,
		class M,
		class T
	>
	class MajaGameEngine_:	public MajaGameLoader	<O,R,T>,
				public MajaGameMenu	<O,I,M>,
				public MajaGameWorld	<O,I,S>
	{
		protected:
		
		//! the moment list is a list that could hang some O in real-time,
		//! so it could be used for temporary or extraordinary objects.
		const MAJACONTAINER::MajaList<O>&	moment_list;

		//! a simple camera_list: for now only the first camera is used.
		MAJACONTAINER::MajaList<O>		camera_list;
		
		//! the game state (cf enum def above)
		MajaGameState				game_state;

		//! the living state is a sub state of gamestate:living_state.
		//! cf enum def above
		MajaLivingState				living_state;

		//! simple count of item (may won't be used anymore)
		muint					item_nb;
		
		MajaGameEngine_( mvoid):	MajaGameLoader<O,R,T>(),
 						MajaGameMenu<O,I,M>(),
						MajaGameWorld<O,I,S>(),
						moment_list( MajaGameEngine_::moment)
		{
			Init();
		}

		~MajaGameEngine_(){
			;
		}

		public:

		// note: there are some bad dones here as AddCamera or SetRenderingEngine...
		
		//! running the game engine: this is the core game management
		//! call it whenever you want the game to be updated.
		virtual mcom Run( mvoid)= 0;

		virtual mcom Unrun( mvoid)= 0;

		virtual mvoid SetInteractiveEngine( I& e)= 0;

		virtual mvoid SetRenderingEngine( R& e)= 0;

		virtual mvoid SetMenuEngine( M& e)= 0;

		virtual muint GetItemNb( mvoid)= 0;

		virtual MajaGameState GetGameState( mvoid)= 0;

		virtual MajaLivingState GetLivingState( mvoid)= 0;

		virtual mvoid AddCamera( O& cam, MajaObjectType type)= 0;

		protected:

		mcom GameStart( mvoid){			
			
			graphic_context= r_engine->GetDrawable().CreateContext();			
			
			if( !graphic_context){
				std::cerr << std::endl << "MajaGameEngine: ERROR while creating context.";
				return mcom_abnormal_error;
			}
			r_engine->GetDrawable().Enter( graphic_context);
			r_engine->GetDrawable().Leave();
			
			return mcom_goto_next;
		}

		mcom GameLoad( mvoid){
			std::cerr << std::endl << "MajaGameEngine: loading...";
			MajaGameLoader<O,R,T>::Run();
			MatchGameWorld();
			std::cerr << std::endl << "MajaGameEngine: computing...";
			MajaGameWorld<O,I,S>::PreRun();
			r_engine->Compute();

			return mcom_goto_next;
		}

		mcom GameLive( mvoid){
			mcom	ans= mcom_nothing;
			
			if( living_state&living_state_menuing){
				ans= MajaGameMenu<O,I,M>::Run();
				if( ans&mcom_exit_program)
					return mcom_exit_program;
				else if( ans&mcom_goto_next){
					std::cerr << std::endl << "MajaGameEngine: leaving menu for game.";
					r_engine->SetUseMenu( mfalse);
					r_engine->SetUseHud( mtrue);
					living_state-=living_state_menuing;
					living_state|=living_state_huding;
				}
			}
			if( living_state&living_state_living){
				ans= MajaGameWorld<O,I,S>::Run();
				
				static mbool first= mtrue;

				if( first){
					r_engine->GetCamera().SetMode( camera_mode_free|camera_mode_user);
					r_engine->GetCamera().SetGravitySubmission( mtrue);
					first=!first;
				}
			}
			else if( living_state&living_state_pause){
				;
			}

			return mcom_no_change;
		}

		mvoid Init( mvoid){
			game_state= game_state_starting;
			living_state= living_state_off;
			item_nb= 0;
		}

		mvoid SetLoaderList( MAJACONTAINER::MajaList<O>& l){
			loader_list= l;
		}

		mvoid SetWorldList( MAJACONTAINER::MajaList<O>& l){
			world_list= l;
		}

		mvoid SetMenuList( MAJACONTAINER::MajaList<O>& l){
			menu_list= m;
		}

		mint ProceedNextState( mvoid){
			Check();
			
			return game_state;

		}

		mint ProceedPreviousState( mvoid){
			Check();
			
			return game_state;
		}

		//! Check manage the inner work to do between a game state change.
		//! note: when starting to loading, the game engine must have some things in
		//! 	its loading list, and must have at least one camera. This behave
		//!	could change in the future.
		mbool Check( mvoid){
			if( game_state== game_state_starting){
				if( loading_list.GetNumber()>0 && camera_list.GetNumber()>0){
					game_state= game_state_loading;
					
					return mtrue;
				}
				if( loading_list.GetNumber()==0) 
					std::cerr << "no item !";
				if( camera_list.GetNumber()==0)
					std::cerr << "no camera !";
			}
			else if( game_state== game_state_loading){
				game_state= game_state_living;
				living_state|= living_state_menuing;
				living_state|= living_state_living;
					
				return mtrue;
			}
	
			return mfalse;
		}

		//! match game world make the match between what's loaded and what will be
		//! refered in the game world.
		mvoid MatchGameWorld( mvoid){
			MAJACONTAINER::MajaWidget<O> *tmp= loading_list.GetFirst();
			MAJACONTAINER::MajaWidget<O> *t= mnull;
			std::list<MajaObjectType>::iterator i= type_list.begin();
			
			if( !tmp)
				return;
			
			do{
				t= tmp;
				AddWorldEntity( *t->GetObject(), *i);
				tmp=tmp->GetNext();
				loading_list.RemoveWidget( t);
				i++;
				type_list.pop_front();
			}while( tmp);

			AddWorldEntity( *camera_list.GetFirst()->GetObject(), object_type_camera|object_type_player);
		}
	};

	template 
	<
		class O, 
		class I, 
		class S,
		class R,
		class M,
		class T
	>
	class MajaGameEngine:	public MajaGameEngine_<O,I,S,R,M,T>
	{
		public:
		
		MajaGameEngine( mvoid):	MajaGameEngine_<O,I,S,R,M,T>()
		{
			;
		}

		~MajaGameEngine(){
			;
		}

		mcom Run( mvoid){
			mcom ans= mcom_nothing;
			
			if( game_state== game_state_starting)
				ans= GameStart();
			else if( game_state== game_state_loading)
				ans= GameLoad();
			else if( game_state== game_state_living)
				ans= GameLive();

			if( ans&mcom_goto_next){
				std::cerr << std::endl << "MajaGameEngine: proceeding to next state...";
				ProceedNextState();
			}
			else if( ans&mcom_goto_prev){
				std::cerr << std::endl << "MajaGameEngine: proceeding to previous state...";
				ProceedPreviousState();
			}
			
			if( ans&mcom_abnormal_error)
				;
			if( ans&mcom_exit_program)
				return mcom_exit_program;
			
			return mcom_no_change;
		}

		mcom Unrun( mvoid){
			if( game_state== game_state_loading)
				MajaGameLoader<O,R,T>::Unrun();
			if( game_state== game_state_living)
				MajaGameWorld<O,I,S>::Unrun();

			return 1;
		}

		mvoid SetInteractiveEngine( I& e){
			MajaGameMenu<O,I,M>::i_engine= &e;
			MajaGameWorld<O,I,S>::i_engine= &e;
		}

		mvoid SetRenderingEngine( R& e){
			r_engine= &e;
		}

		mvoid SetMenuEngine( M& e){
			MajaGameMenu<O,I,M>::m_engine= &e;
		}

		muint GetItemNb( mvoid){
			item_nb= 0;
			item_nb+= living_list.GetNumber();
			item_nb+= dead_list.GetNumber();
			item_nb+= tool_list.GetNumber();
			item_nb+= camera_list.GetNumber();

			return item_nb;
		}

		MajaGameState GetGameState( mvoid){
			return game_state;
		}

		MajaLivingState GetLivingState( mvoid){
			return living_state;
		}

		mvoid AddCamera( O& cam, MajaObjectType type){
			camera_list.AddObject( &cam);
			if( type&object_type_player){
				std::cerr << std::endl << "MajaGameEngine: camera is the player.";
				//AddWorldEntity( cam, type);
			}
			r_engine->SetCamera( &cam);
		}
	};
}
}
#endif
