/*
	majagconf.h
	ce fichier est une partie de Maja - this is a part of Maja
	Configuration OpenGL
	OpenGL configuration

	$Id: $

	$Log: $

	begin:			Sat Dec 14 2002
	copyright:		(C) 2000,2001,2002,2003 by Jean-Dominique Frattini
	email:			jd.frattini@tuxfamily.org	
*/

/************************************************************************************************
 *												*
 *	Ce programme est un logiciel libre; vous pouvez le redistribuer et/ou le modifier	*
 *	sous les termes de la GNU General Public License comme publie par			*
 *	la Free Software Foundation; soit en utilisant la verstion 2 de cette license, ou	*
 *	(si vous le voulez) toute version ultrieure.						*
 *	-											*
 *	This program is free software; you can redistribute it and/or modify			*
 *	it under the terms of the GNU General Public License as published by			*
 *	the Free Software Foundation; either version 2 of the License, or			*
 *	(at your option) any later version.							*
 *												*
 ************************************************************************************************/

#ifndef MAJAGCONF__H
#define MAJAGCONF__H


#ifndef WIN32
#include <GL/glx.h>
#endif

#include "../../tools/maths/majamaths.h"

namespace maja
{
namespace MAJAGRAPHIC
{
	using MAJAMATHS::MajaQuaternion;

	/*!info
	 * these are the new ways i have keep for maja
	 * enumeration definitions. So, no trigrams, no
	 * bad understanding, any value name all the 2
	 * words after the MAJA enum name.
	 */

	/*!
	 * def:	OpenGL shading model
	 * state: safe
	 */
	enum MajaShadeModel{
		shade_model_smooth=	GL_SMOOTH,
		shade_model_flat=	GL_FLAT
	};

	/*!
	 * def:	OpenGL smooth rendering hints
	 * state: safe
	 */
	enum MajaSmoothLevel{
		smooth_level_nicest=	GL_NICEST,
		smooth_level_fastest=	GL_FASTEST,
		smooth_level_auto=	GL_DONT_CARE
	};

	/*!
	 * def: OpenGL buffers use
	 * inf:	other buffers may be needed.
	 * 	this is not a real enum type, as we want to make
	 * 	some bit operations using them.
	 * state: safe
	 */
	typedef GLenum MajaUseBuffer;
	const MajaUseBuffer	use_buffer_color=	GL_COLOR_BUFFER_BIT;
	const MajaUseBuffer	use_buffer_depth=	GL_DEPTH_BUFFER_BIT;
	const MajaUseBuffer	use_buffer_stencil=	GL_STENCIL_BUFFER_BIT;

	enum MajaCullFace{
		cull_face_front=GL_FRONT,
		cull_face_back=	GL_BACK,
		cull_face_all=	GL_FRONT_AND_BACK
	};

	enum MajaFrontFace{
		front_face_ccw=	GL_CCW,
		front_face_cw=	GL_CW
	};

	/*!
	 * def:	simple OpenGL graphic configuration
	 * inf:	should assume it's only a singleton
	 * 	all the member functions use referencies as parameters and as return
	 * 	values, if exist.
	 * 	by default, depth tests, blending and texturing are set, background
	 * 	color is black, shade model is smooth, smooth level is to nicest (so
	 * 	the 'slowest' opengl rendering mode) and maja uses color, depth and
	 * 	stencil buffers.
	 * state: in development
	 */
	class MajaGConf
	{
		//! color of the OpenGL screen (so background)
		static MajaQuaternion	background_color;

		//! shading model
		static MajaShadeModel	shade_model;

		//! use of depth tests
		static bool		use_hardware_depth_test;

		//! use of blending
		static bool		use_blending;

		//! level of smoothness desired
		static MajaSmoothLevel	smooth_level;

		//! use of texturing
		static bool		use_texture;

		//! use of the buffers
		static MajaUseBuffer	use_buffer;

		//! use of lighting
		static bool		use_lighting;

		static MajaCullFace	cull_face;

		static MajaFrontFace	front_face;

		//polygon mode (as cull_face)

		public:

		MajaGConf( void){
			use_hardware_depth_test=	false;
			use_blending=			false;
			use_texture=			false;
			use_lighting=			false;
		}

		~MajaGConf( void){
			;
		}

		void Init( void){
			background_color= 0;
			SetBackgroundColor( background_color);
//			SetUseHardwareDepthTest( true);
//			SetUseBlending( false);
//			SetUseTexture( true);
			SetShadeModel( shade_model_smooth);
			SetSmoothLevel( smooth_level_nicest);
			SetBufferUse( use_buffer_color|use_buffer_depth|use_buffer_stencil);
			SetUseLighting( true);
			SetCullFace( cull_face_back);
			SetFrontFace( front_face);
			glPushClientAttrib(	GL_ENABLE_BIT | 
						GL_COLOR_BUFFER_BIT | 
						GL_DEPTH_BUFFER_BIT | 
						GL_LIGHTING_BIT | 
						GL_TEXTURE_BIT);
			glDepthRange( 0, 100);
		}

		//! dangerous way if it's not in the correct matrix element...		
		void Reset( void){
			glPopClientAttrib();
		}

		void SetBackgroundColor( MajaQuaternion& color){
			background_color= color;
			glClearColor( color.x, color.y, color.z, color.w);
		}

		MajaQuaternion& GetBackgroundColor( void){
			return background_color;
		}

		void SetShadeModel( MajaShadeModel model){
			glShadeModel( shade_model=model);
		}

		MajaShadeModel& GetShadeModel( void){
			return shade_model;
		}

		void SetUseHardwareDepthTest( bool use){
			use_hardware_depth_test= use;
			if( use_hardware_depth_test)
				glEnable( GL_DEPTH_TEST);
			else
				glDisable( GL_DEPTH_TEST);
		}

		bool& GetUseHardwareDepthTest( void){
			if( !glIsEnabled( GL_DEPTH_TEST))
				return use_hardware_depth_test= false;
			return use_hardware_depth_test;
		}

		void SetUseBlending( bool use){
			use_blending= use;
			if( use){
				glEnable( GL_BLEND);
				glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			}
			else
				glDisable( GL_BLEND);
		}

		bool& GetUseBlending( void){
			return use_blending;
		}

		void SetSmoothLevel( MajaSmoothLevel level){
			glHint( GL_POLYGON_SMOOTH_HINT, smooth_level= level);
		}

		MajaSmoothLevel& GetSmoothLevel( void){
			return smooth_level;
		}

		void SetUseTexture( bool use){
			use_texture= use;
			if( use_texture)
				glEnable( GL_TEXTURE_2D);
			else
				glDisable( GL_TEXTURE_2D);
		}

		bool& GetUseTexture( void){
			return use_texture;
		}

		void SetBufferUse( MajaUseBuffer use){
			use_buffer= use;
		}

		void ClearBuffers( void){
			glClear( use_buffer);
		}

		MajaUseBuffer& GetBufferUse( void){
			return use_buffer;
		}

		void SetUseLighting( bool use){
			use_lighting= use;
			if( use_lighting)
				glEnable( GL_LIGHTING);
			else
				glDisable( GL_LIGHTING);
		}

		bool& GetUseLighting( void){
			return use_lighting;
		}

		void SetCullFace( MajaCullFace mode){
			cull_face= mode;
			glCullFace( cull_face);
		}

		MajaCullFace& GetCullFace( void){
			return cull_face;
		}

		void SetFrontFace( MajaFrontFace face){
			front_face= face;
			glFrontFace( front_face);
		}

		MajaFrontFace& GetFrontFace( void){
			return front_face;
		}
	};
}//namespace MAJAGRAPHIC
}
#endif
