/*
	majagengine.h
	ce fichier est une partie de Maja - this is a part of Maja
	moteur graphique de rendu "optimis" pour geforces.
	graphic engine of "optimized" rendering for geforces.

	$Id: $

	$Log: $

	begin:			Wed Jun 19 2002
	copyright:		(C) 2000,2001,2002,2003 by Jean-Dominique Frattini
	email:			jd.frattini@tuxfamily.org	
*/

/************************************************************************************************
 *												*
 *	Ce programme est un logiciel libre; vous pouvez le redistribuer et/ou le modifier	*
 *	sous les termes de la GNU General Public License comme publie par			*
 *	la Free Software Foundation; soit en utilisant la verstion 2 de cette license, ou	*
 *	(si vous le voulez) toute version ultrieure.						*
 *	-											*
 *	This program is free software; you can redistribute it and/or modify			*
 *	it under the terms of the GNU General Public License as published by			*
 *	the Free Software Foundation; either version 2 of the License, or			*
 *	(at your option) any later version.							*
 *												*
 ************************************************************************************************/

#ifndef __MAJA_G_ENGINE_H
#define __MAJA_G_ENGINE_H

#include "graphext.h"
#include "gbase.h"
#include "../modeler/majamesh.h"
#include "majagmaterial.h"
#include "majagtext.h"
#include "majagdata.h"
#include "majagmesh.h"
#include "majagintegrator.h"
#include "graphictree.h"
#include "../../tools/containers/majalist.h"
#include "majagconf.h"

#include "../../core/majamemmgr.h"

#include <iostream>

namespace maja
{

namespace MAJAGRAPHIC{

/*!
	the graphic engine is intended to render the world game.
	It depends only on the objects who has got a MajaGMesh.
	So, you need one or more in order to render with it.
	generally, there will be only one instance of this type.
	
	infos:
	gbase.h defines some OpenGL rendering technics that could stands the engine.
	not all is supported, so check this file for more info.
	
	you could draw here with vertex arrays only. for display list, cf majadisplay.h
	
	This engine is fair optimized for NVidia VAR fast renderings, but it could
	switch down to some other technics as CVA (assumes that all data
	won't be transformed anyway), or lesser as simple VA.
	
	status: in development, quite useable

	version : 0.1.3
*/
class MajaGraphicEngine: public MajaGEntity
{
	//! the name of an instance of this class
	mchar						*maja_internal_name;
	
	protected:
	
	//! the root of the graphic tree
	MajaGRoot					graphic_root;
	
	//! a list of MajaGMesh to be rendered
	MAJACONTAINER::MajaList<MajaGMesh>		meshes_list;
	
	//! leaves of the tree: only internal use (may be listed too)
	MajaGDataTree					*leaves;
	
	//! breanches: only internal use
	MajaGParam					*branches;
	
	// maybe in _MajaRenderInfo
	
	//! nb of leaves
	GLuint						total_leaves_count;	// care of bugs due to limitation in bits...
	
	//! nb of branch
	GLuint						total_branch_counter;
	
	//! not null if any bad leave and represent the nb in the list
	GLuint 						*dum;
	
	//! count the dummy leaves ref to *dum
	GLuint 						dum_count;
	
	//! count the meshes in the list
	GLuint						meshes_count;
	
	//! mtrue if data are locked into AGP memory (it's the cva drawing method)
	mbool						data_locked;
	
	//! some infos about what's rendered
	struct _MajaRenderInfo
	{
		//! the total number of faces in the actual renderer
		GLuint	faces_nb;
		
		//! an array for indexes the vertices
		GLuint	*vert_index;
		
		~_MajaRenderInfo( mvoid)
		{
			if( vert_index) delete vert_index;
		}
	}						render_info;

	MajaGConf					gl_conf;

	MajaMemoryManager<mfloat>			*mem_mgr;

	mfloat 						*agp_memory_pointer;
	
	public:
	
	MajaGraphicEngine();
	
	~MajaGraphicEngine();
	
	//! render the whole tree
	inline GLuint Render( mvoid)
	{
	//	GLuint i=0;
		static mbool first= mtrue;

		if( first){
			gl_conf.SetUseHardwareDepthTest( mtrue);
			first= mfalse;
		}
		
//		for( i=0; i<graphic_root.branch_nb; i++)
//			graphic_root.graphic_branch[ i ].Render();
			
		graphic_root.Render();
		
		return 1;
	}
	
	//! adding meshes: only add mesh one by one, so nb=1
	mvoid AddMesh( MajaGMesh *m, muint nb);
	
	//! call it when you have all your meshes in and ready to render
	GLuint Compute( mvoid);
	
	protected:
	
	//! make first OpenGL preparations
	GLuint ComputeStarts( mvoid);
	
	//! internal use (iu): create leaves
	GLuint ComputeLeaves( mvoid);
	
	//! iu: create branches
	GLuint ComputeBranches( mvoid);
	
	//! iu: link branches with leaves
	//! for now: only one leave by branch
	//! later: a branch could have 1..* child branches that could be leaves
	GLuint ComputeLinks( mvoid);
	
	//! iu: create the entire graphic tree
	GLuint ComputeTree( mvoid);
	
	/*!
		iu: terminate all computes: extensions, arrays...
		what is done here:
		check what is known supported
	*/
	GLuint ComputeEnds( mvoid);
	
	public:
	
	//! call it to check the integrity of all the infos
	mbool Check( mvoid);
	
	//! check some OpenGL states to have a correct init of the renderer
	mbool CheckStates( mvoid);

	inline mvoid SetMemoryManager( MajaMemoryManager<mfloat>& ptr){
		mem_mgr= &ptr;
	}
	
	protected:
	
	inline mvoid EnableNormalArrays( mvoid){
		glEnableClientState( GL_VERTEX_ARRAY);
		glEnableClientState( GL_NORMAL_ARRAY);
		glEnableClientState( GL_COLOR_ARRAY);
		glEnableClientState( GL_TEXTURE_COORD_ARRAY);
	}
	
	inline mvoid EnableExtArrays( mvoid){
		glEnableClientState( GL_VERTEX_ARRAY_EXT);
		glEnableClientState( GL_NORMAL_ARRAY_EXT);
		glEnableClientState( GL_COLOR_ARRAY_EXT);
		glEnableClientState( GL_TEXTURE_COORD_ARRAY_EXT);
		
		if( gl_ext->GetNvDraw())
			glEnableClientState( GL_VERTEX_ARRAY_RANGE_NV);
	}
	
	inline mvoid DisableNormalArrays( mvoid){
		glDisableClientState( GL_VERTEX_ARRAY);
		glDisableClientState( GL_NORMAL_ARRAY);
		glDisableClientState( GL_COLOR_ARRAY);
		glDisableClientState( GL_TEXTURE_COORD_ARRAY);
	}
	
	inline mvoid DisableExtArrays( mvoid){
		glDisableClientState( GL_VERTEX_ARRAY_EXT);
		glDisableClientState( GL_NORMAL_ARRAY_EXT);
		glDisableClientState( GL_COLOR_ARRAY_EXT);
		glDisableClientState( GL_TEXTURE_COORD_ARRAY_EXT);
		
		// this could not be disabled: sometimes, nvdraw is mtrue first, & change while
		// the program is running. So, when ending, it's like it's never been initialized.
		if( gl_ext->GetNvDraw())
			glDisableClientState( GL_VERTEX_ARRAY_RANGE_NV);
	}
};
}// namespace MAJAGRAPHIC
}
#endif
