/*
	majaobject.h
	ce fichier est une partie de Maja - this is a part of Maja
	objet-mre de la partie gestion de jeux
	mother-object of the game management part
	
	$Id: $

	$Log: $

	begin:			Sep 10 2000
	copyright:		(C) 2000,2001,2002,2003 by Jean-Dominique Frattini
	email:			jd.frattini@tuxfamily.org	
*/

/************************************************************************************************
 *												*
 *	Ce programme est un logiciel libre; vous pouvez le redistribuer et/ou le modifier	*
 *	sous les termes de la GNU General Public License comme publie par			*
 *	la Free Software Foundation; soit en utilisant la verstion 2 de cette license, ou	*
 *	(si vous le voulez) toute version ultrieure.						*
 *	-											*
 *	This program is free software; you can redistribute it and/or modify			*
 *	it under the terms of the GNU General Public License as published by			*
 *	the Free Software Foundation; either version 2 of the License, or			*
 *	(at your option) any later version.							*
 *												*
 ************************************************************************************************/

#ifndef MAJA_OBJECT_H
#define MAJA_OBJECT_H

#include <stdlib.h>
#include <typeinfo>
#include <string>
#include "../../common/majatypes.h"
#include "../../tools/maths/geometry.h"
#include "../../tools/maths/majapoly.h"
#include "../modeler/modelio.h"			//MajaObjectFactory will avoid that
#include "../../core/majaclock.h"

namespace maja
{

static const mdouble GRAVITY= 		9.8;

/*!
	MajaObject is the mother class of all the graphic objects of Maja.
	This is the main object of the polymorphic hierarchy.
	This contains most functionnality common to the hierarchy objects.
	This is the most simple displayable object.

	status: useable.

	infos: this is one of the older classes of the project, but has mutate
		some and will change.
*/
class MajaObject
{
	protected:
		
	//! position of the object
	MajaVector 			pos;	
	
	//! angle around the 3 Opengl axes: angle.x is around the y axis, angle.y is around the x axis
	MajaVector			angle;
	
	//! volume of the cubic 8 points form
	MajaVector 			volpts[8];
	
	//! maximal sizez of the 3 dimensions
	MajaVector 			size;
	
	//! name
	std::string 			name;
	
	//! 3d model module
	MajaModule			*mod;
	
	std::string			model_file;
	
	mdouble				mass;
	
	mbool				gravity_submission;

	MajaVector			acceleration;
	MajaVector			speed;

	//! internal clock
	MajaClock    	inner_clk;

	public:
	
	//! Creating a MajaObject 
	MajaObject( mvoid);
	
	//protected: maybe one day
	
	MajaObject( const MajaObject& mo);

	virtual ~MajaObject( mvoid);

	//!
	//! Forming, drawing the object:
	//!
	
	//! load a 3d model from an input file (need the directory too)
	virtual mbool Load();

	mvoid SetModelFile( std::string m_file);

	//! free if the 3D model exists
	mvoid FreeModel( mvoid);
	
	//! this new virtual function is dedicated to all the objects who
	//! do not act in the world, but in the scene decoration
	virtual mvoid Presence( mvoid);
	
	//! get the 3D model
	inline MajaModule& GetModel( mvoid){ return *mod;}
	
	//! set it (beware of any previous allocation)
	inline mvoid SetModel( MajaModule& s){ mod= &s;}
	
	//!
	//! Transformations:
	//!
	
	mvoid SetOrientation( MajaVector a){
		SetAngles( a);
	}

	mvoid SetAngles( MajaVector new_angle, mbool active= mtrue);

	MajaVector& GetOrientation(){
		return GetAngles();
	}
	
	inline MajaVector& GetAngles( mvoid){
		return angle;
	}
	
	//! return the position
	inline MajaVector& GetPos( mvoid){
		return pos;
	}
	
	//! set the new position
	inline mvoid SetPos( MajaVector p){
		pos= p;
	}

	inline mvoid SetPos( mreal x, mreal y, mreal z){
		pos.x=x;
		pos.y=y;
		pos.z=z;
	}
	
	//! Set the size to the given parameter
	inline mvoid SetSize( MajaVector s){ 
		size=s;
	}
	
	//! Get the size
	inline MajaVector& GetSize( mvoid){
		return size;
	}

	//!
	//! Tools:
	//!
	
	//! Define volpts member due to position, sizes
	virtual mvoid DefineVolume( mvoid);

	mvoid SetGravitySubmission( mbool s){ 
		gravity_submission= s;
	}
	
	//! these private methods aren't set to be used by any user, just for internal
	//! compliance. Is the drawing of the object set from its center or by the lower
	//! angle ? The normal way is to set by the center
	
	private:
	
	// the volume is always defined by the position on the object (its center by default).
	// so all the calculations will use this position.
	mvoid DefineVolume( real px, real py, real pz, real qx, real qy, real qz);
	
	//! considere that the center of the object is the origine
	mvoid DefineVolumeFromCenter( mvoid);
	
	//! consider that the lower minus angle is the origine
	mvoid DefineVolumeFromAngle ( mvoid);   //! isn't used now
	
	mvoid RotateVolume( mvoid);

	mvoid SetMass( mdouble m){
		mass= m;
	}
	
	public:

	//!
	//! Actions:
	//!
	
	//! put the object on an arbitrary ground.
	//! should be able to use a MajaObject descendant
	//! but not known from now     	
	virtual mint ObjectOnFloor  ( const MajaObject& floor);

	mvoid SetAcceleration( MajaVector a){
		acceleration= a;
	}

	MajaVector& GetAcceleration( mvoid){
		return acceleration;
	}

	MajaVector& GetSpeed( mvoid){
		return speed;
	}

	// will need to redefine the acceleration related
	// to the new speed vector.
	mvoid SetSpeed( MajaVector s){
		speed= s;
	}

	virtual MajaVector& CalculateSpeed( mint sens);

	virtual MajaVector& Accelerate( mfloat k);
	virtual MajaVector& Decelerate( mfloat k);

	mbool GetGravitySubmission( mvoid){
		return gravity_submission;
	}
	
	//!
	//! Naming  methods:
	//!
	
	//! Set the name:
	mvoid SetName( std::string& n);
	
	//! Get the name:
	//std::string& GetName( mvoid);
	mchar*  GetName( mvoid);
	
	//! Check if the name matches with the seeked one,
	//! useful for list containers to seek one named object
	inline mbool CheckName( std::string& n);
	
	//!
	//! Operators:
	//!
	
	//! check if the 2 objects are the same or not
	mbool operator==( MajaObject& src);

	MajaVector* GetVolPts( mvoid){
		return volpts;
	}
};

}
#endif
