/*
	majavector.h
	ce fichier est une partie de Maja - this is a part of Maja
	une classe pour les vecteurs 3D
	a class for 3D vectors

	$Id: $

	$Log: $

	begin:			Sun Aug 11 2002
	copyright:		(C) 2000,2001,2002,2003 by Jean-Dominique Frattini
	email:			jd.frattini@tuxfamily.org	
*/

/************************************************************************************************
 *												*
 *	Ce programme est un logiciel libre; vous pouvez le redistribuer et/ou le modifier	*
 *	sous les termes de la GNU General Public License comme publie par			*
 *	la Free Software Foundation; soit en utilisant la verstion 2 de cette license, ou	*
 *	(si vous le voulez) toute version ultrieure.						*
 *	-											*
 *	This program is free software; you can redistribute it and/or modify			*
 *	it under the terms of the GNU General Public License as published by			*
 *	the Free Software Foundation; either version 2 of the License, or			*
 *	(at your option) any later version.							*
 *												*
 ************************************************************************************************/

#ifndef MAJA_VECTOR_H
#define MAJA_VECTOR_H

#include "majascalar.h"

namespace maja
{
namespace MAJAMATHS{

//! size of a vector; default way is 3 dimensions
//! you should not change this value in run-time
#define MAJA_VECTOR_SIZE	3

/*!
	A class that hangs 3D geometric vector.
	But the size could change by manipulating the value
	of MAJA_VECTOR_SIZE which is 3 dimensionned by default.
	
	status: most tests passed, is pretty usable
		miss some operators(++,--)

	todo: add new functions members
		Rotate( MajaVector& axis, _scalar angle) which rotate the vector around the axis to angle radians.
*/
class MajaVector: virtual public MajaScalar
{
	protected:
	
	//! the length of the vector
	mdouble		magnitude;
	
	public:
	
	//! 3 references linked to specific array element.
	//! this is in user access to avoid methods calls overhead.
	_scalar&		x;
	_scalar&		y;
	_scalar&		z;
	
	//! default constructor
	MajaVector( mvoid):	MajaScalar( MAJA_VECTOR_SIZE),
				x( *scalar_ptr),
				y( *(scalar_ptr+1)),
				z( *(scalar_ptr+2))
	{
		magnitude= 0;
	}
	
	//! when construct with it, your vector has the same values
	//! for all its 3 components (x, y & z)
	MajaVector( _scalar v):	MajaScalar( MAJA_VECTOR_SIZE),
				x( *scalar_ptr),
				y( *(scalar_ptr+1)),
				z( *(scalar_ptr+2))
	{
		scalar_ptr[0]= v;
		scalar_ptr[1]= v;
		scalar_ptr[2]= v;
		CalculateMagnitude();
	}
	
	//! most normal way of instanciate a vector
	MajaVector( _scalar v1, _scalar v2, _scalar v3): 	MajaScalar( MAJA_VECTOR_SIZE),
								x( *scalar_ptr),
								y( *(scalar_ptr+1)),
								z( *(scalar_ptr+2))
	{
		scalar_ptr[0]= v1;
		scalar_ptr[1]= v2;
		scalar_ptr[2]= v3;
		CalculateMagnitude();
	}
	
	//! copy constructor
	MajaVector( const MajaVector& v):	MajaScalar( MAJA_VECTOR_SIZE),
						x( *scalar_ptr),
						y( *(scalar_ptr+1)),
						z( *(scalar_ptr+2))
	{
		scalar_ptr[0]= v.scalar_ptr[0];
		scalar_ptr[1]= v.scalar_ptr[1];
		scalar_ptr[2]= v.scalar_ptr[2];
		magnitude= v.magnitude;
	}

	~MajaVector( mvoid)
	{
		Free();
	}
	
	//! set / get the different values of its components
	virtual inline mvoid SetXValue( _scalar val)
	{ 
		x= val; 
		CalculateMagnitude();
	}
	
	virtual inline _scalar	GetXValue( mvoid)
	{ 
		return x;
	}
	
	virtual inline mvoid SetYValue( _scalar val)
	{ 
		y= val; 
		CalculateMagnitude();
	}
	
	virtual inline _scalar	GetYValue( mvoid)
	{ 
		return y;
	}
	
	virtual inline mvoid SetZValue( _scalar val)
	{ 
		z= val; 
		CalculateMagnitude();
	}
	
	virtual inline _scalar	GetZValue( mvoid)
	{ 
		return z;
	}
	
	virtual inline mvoid Print( mvoid)
	{ 
		std::cout << std::endl << x << " ; " << y << " ; " << z << std::flush;
	}	
	
	//! calculate the magnitude in relation with its value
	virtual inline _scalar CalculateMagnitude( mvoid)
	{
		return magnitude= sqrt( x*x+y*y+z*z);
	}
	
	//! normalize of vector with a special length
	//! normal way of normalization is to set the lenght to 1
	virtual inline mvoid Normalize( _scalar s=1)
	{
		CalculateMagnitude();
		mdouble v=s/magnitude;
		x*=v;
		y*=v;
		z*=v;
		magnitude= s;
	}
	
	//! return the dot product of 2 vectors
	virtual inline _scalar DotProduct( const MajaVector& v)
	{
		_scalar p;
		
		p= x*v.x+y*v.y+z*v.z;
		
		return p;
	}
	
	//! return the cross product of 2 vectors
	virtual inline MajaVector CrossProduct( const MajaVector& v)
	{
		MajaVector	p;
		
		p.x= y*v.z-z*v.y;		
		p.y= z*v.x-x*v.z;		
		p.z= x*v.y-y*v.x;
		
		return p;
	}
	
	//! return mtrue if this is the null vector, otherwise mfalse
	virtual inline mbool IsNull( mvoid)
	{
		if( x==y && x==z && x==0)
			return mtrue;
		
		return mfalse;
	}
	
	//! return the angle, in radians, made between
	//! the 2 vectors, in CCW normal math system
	//! orientation, made btw this vector & the specified
	virtual inline _scalar GetAngle( const MajaVector& v)
	{
		CalculateMagnitude();
		
		return acos( DotProduct( v)/(magnitude*v.magnitude));
	}
	
	//! get the angle around the y axis. the vectors are restreinct
	//! into 2D vector with x & z only
	virtual inline _scalar GetYAngle( const MajaVector& v)
	{
		return acos( (x*v.x+z*v.z)/(sqrt( x*x+z*z)*sqrt( v.x*v.x+v.z*v.z)));
	}

	//! TODO: MajaVector GetAngle( const MajaVector&) to return the angles btw all the axis
	
	//
	//! calculation and affectation operators:
	//
	virtual inline MajaVector operator+		( const MajaVector& m)
	{
		MajaVector p;
		
		p.x= x+ m.x;
		p.y= y+ m.y;
		p.z= z+ m.z;
		
		return p;
	}
	
	//! substract 2 vectors
	virtual inline MajaVector operator-		( const MajaVector& m)
	{
		MajaVector p;
		
		p.x= x- m.x;
		p.y= y- m.y;
		p.z= z- m.z;
		
		return p;
	}
	
	//! multiply a vector with a _scalar value (any float compatible)
	virtual inline MajaVector operator*		( const _scalar v)
	{
		MajaVector p;

		p.x= x*v;
		p.y= y*v;
		p.z= z*v;
		
		return p;
	}
	
	//! this applies the matrix multiplication restricted to columns
	virtual inline MajaVector operator*		( const MajaVector& m)
	{
		MajaVector p;
		
		_scalar v= m.x+m.y+m.z;
		
		p.x= x*v;
		p.y= y*v;
		p.z= z*v;
		
		return p;
	}
	
	//! this applies the matrix division restricted to columns
	virtual inline MajaVector operator/		( const MajaVector& m)
	{
		MajaVector p;
		
		_scalar v= 1/(m.x+m.y+m.z);
		
		p.x= x*v;
		p.y= y*v;
		p.z= z*v;		
		
		return p;
	}
	
	//! this divides the vector by a real
	virtual inline MajaVector operator/		( const _scalar m)
	{
		MajaVector p;
		
		_scalar v= 1/m;
		
		p.x= x*v;
		p.y= y*v;
		p.z= z*v;		
		
		return p;
	}
	
	//! Add a vector to this one
	virtual inline MajaVector operator+=		( const MajaVector& m)
	{
		unsigned int i;
		
		for( i=0; i<MAJA_VECTOR_SIZE; i++)
			scalar_ptr[i]+= m.scalar_ptr[i];
			
		return *this;
	}
	
	//! substract a vector to this one
	virtual inline MajaVector operator-=		( const MajaVector& m)
	{
		unsigned int i;
		
		for( i=0; i<MAJA_VECTOR_SIZE; i++)
			scalar_ptr[i]-= m.scalar_ptr[i];
			
		return *this;
	}
	
	//! multiply each by each any values by this vector
	virtual inline MajaVector operator*=		( const MajaVector& m)
	{
		unsigned int i;
		
		for( i=0; i<MAJA_VECTOR_SIZE; i++)
			scalar_ptr[i]*= m.scalar_ptr[i];
			
		return *this;
	}
	
	virtual inline MajaVector operator/=		( const MajaVector& m)
	{
		unsigned int i;
		
		for( i=0; i<MAJA_VECTOR_SIZE; i++)
			scalar_ptr[i]/= m.scalar_ptr[i];
			
		return *this;
	}
	
	
	virtual inline MajaVector operator--		( mvoid)
	{
		muint i;

		for( i=0; i<entity_nb; i++)
			scalar_ptr[i]-= magnitude/(magnitude-1);

		return *this;
	}
	
	virtual inline MajaVector operator++	( mvoid)
	{
		muint i;

		CalculateMagnitude();

		for( i=0; i<entity_nb; i++)
			scalar_ptr[i]+= magnitude/(magnitude+1);

		return *this;
	}
	
	virtual inline MajaVector operator--		( int)
	{
		muint i;

		CalculateMagnitude();

		for( i=0; i<entity_nb; i++)
			scalar_ptr[i]-= magnitude/(magnitude-1);

		return *this;
	}
		
	virtual inline MajaVector operator++	( int)
	{
		muint i;

		for( i=0; i<entity_nb; i++)
			scalar_ptr[i]+= magnitude/(magnitude-1);

		return *this;
	}
		
	//! affectation operator
	virtual inline MajaVector& operator=		( const MajaVector m)
	{
		muint i;
		
		for( i=0; i<MAJA_VECTOR_SIZE; i++)
			scalar_ptr[i]= m.scalar_ptr[i];
			
		return *this;
	}

/*	virtual inline MajaVector& operator =		( const _scalar* v)
	{
		muint i;

		for( i=0; i<MAJA_VECTOR_SIZE; i++)
			scalar_ptr[i]= v[i];

		return *this;
	}
*/	
	//! strict inferior comparison
	virtual inline mbool     operator< 			( const MajaVector& m)
	{
		if( ( x< m.x) && ( y< m.y) && ( z< m.z))
			return mtrue;
		return mfalse;
	}
	
	//! inferior or equal comparison
	virtual inline mbool     operator<= 			( const MajaVector& m)
	{
		if( ( x<= m.x) && ( y<= m.y) && ( z<= m.z))
			return mtrue;
		return mfalse;
	}
	
	//! strict superior comparison
	virtual inline mbool     operator> 			( const MajaVector& m)
	{
		if( ( x> m.x) && ( y> m.y) && ( z> m.z))
			return mtrue;
		return mfalse;
	}
	
	//! superior or equal comparison
	virtual inline mbool     operator>= 			( const MajaVector& m)
	{
		if( ( x>= m.x) && ( y>= m.y) && ( z>= m.z))
			return mtrue;
		return mfalse;
	}
	
	//! equal comparison
	virtual inline mbool     operator==			( const MajaVector& m)
	{
		if( ( x== m.x) && ( y== m.y) && ( z== m.z))
			return mtrue;
		return mfalse;
	}
	
	//! different comparison
	virtual inline mbool     operator!=			( const MajaVector& m)
	{
		if( ( x== m.x) && ( y== m.y) && ( z== m.z))
			return mfalse;
		return mtrue;
	}
};

}//namespace MAJAMATHS
}
#endif
